/*
 * Copyright (c) 2004 Alessandro Rubini. GNU GPL 2 or later
 * Heavily based on drivers/usb/storage/
 *   (c) 1999-2003 Matthew Dharm (mdharm-usb@one-eyed-alien.net)
 *   (c) 2000 David L. Brown, Jr. (usb-storage@davidb.org)
 *   (c) 2003 Alan Stern (stern@rowland.harvard.edu)
 *   (c) 1999 Michael Gee (michael@linuxspecific.com)
 *   (c) 2000 Yggdrasil Computing, Inc.
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/mm.h>
#include <linux/completion.h>
#include <asm/unaligned.h>

#include "scsi.h"
#include <scsi/scsi.h>
#include <scsi/scsi_host.h>

#include "sdm.h"

MODULE_LICENSE("GPL");

struct sdm_dev {
	struct Scsi_Host *host;
	Scsi_Cmnd *srb;
	struct semaphore sema;
	struct completion notify;
	int doexit;
} sdm_dev;

/*
 * Low-level command handling
 */
static void sdm_scsi_lowlevel_command(Scsi_Cmnd *srb)
{
	switch(srb->cmnd[0]) {
		case READ_CAPACITY:
		{
			struct sizeinfo {u32 nblocks; u32 blocksize;} size;
			/* PAGE_SIZE blocks */
			size.nblocks = cpu_to_be32(sdm_size<<(20-PAGE_SHIFT));
			size.blocksize = cpu_to_be32(PAGE_SIZE);
			
			if (srb->request_bufflen > sizeof(size))
				srb->request_bufflen = sizeof(size);
			memcpy(srb->request_buffer, &size, sizeof(size));
			srb->result = SAM_STAT_GOOD;
			break;
		}

		case READ_10:
			if (sdm_verbose > 1)
				printk("read %i bytes (%i blocks), at %i\n",
				       srb->request_bufflen,
				       be16_to_cpu(get_unaligned(
					       (u16*)(srb->cmnd+7))),
				       be32_to_cpu(get_unaligned(
					       (u32*)(srb->cmnd+2))));
#if 0
			if (sd_read(dev, srb->request_buffer, 
				    srb->request_bufflen, 
				    __get_unaligned_be(
					    (u32 *)(srb->cmnd+2))) < 0)
				return DID_ERROR << 16;
#endif
			srb->result = SAM_STAT_GOOD;
			break;

		case WRITE_10:
			if (sdm_verbose > 1)
				printk("write %i bytes (%i blocks), at %i\n",
				       srb->request_bufflen,
				       be16_to_cpu(get_unaligned(
					       (u16*)(srb->cmnd+7))),
				       be32_to_cpu(get_unaligned(
					       (u32*)(srb->cmnd+2))));
#if 0
			if (sd_write(dev, srb->request_buffer, 
				     srb->request_bufflen, 
				     __get_unaligned_be(
					     (u32 *)(srb->cmnd+2))) < 0)
				return DID_ERROR << 16;
#endif
			srb->result = SAM_STAT_GOOD;
			break;
			       
			       
			
		case MODE_SENSE:
			memset(srb->request_buffer, 0, 8);
			if (0) {
				/* this applies to read-only media */
				((unsigned char *)srb->request_buffer)[2]
					|= 0x80;
			}
			/* fall through */
		case TEST_UNIT_READY:
			srb->result = SAM_STAT_GOOD;
			break;
		       
		case ALLOW_MEDIUM_REMOVAL:
			if (srb->cmnd[4] & 1) {
				/* removal is denied */
				try_module_get(THIS_MODULE);
			} else {
				/* removal is allowed */
				module_put(THIS_MODULE);
			}
			srb->result = SAM_STAT_GOOD;
			break;
		default:
			printk("unhandled command %i\n", srb->cmnd[0]);
			srb->result = SAM_STAT_GOOD;
	}
	return;
}


/*
 * Command table, for messages
 */
struct sdm_scsi_cmdname {
	char *name;
	int value;
} sdm_scsi_cmdtable [] = {
	{"TEST_UNIT_READY",       0x00},
	{"REZERO_UNIT",           0x01},
	{"REQUEST_SENSE",         0x03},
	{"FORMAT_UNIT",           0x04},
	{"READ_BLOCK_LIMITS",     0x05},
	{"REASSIGN_BLOCKS",       0x07},
	{"READ_6",                0x08},
	{"WRITE_6",               0x0a},
	{"SEEK_6",                0x0b},
	{"READ_REVERSE",          0x0f},
	{"WRITE_FILEMARKS",       0x10},
	{"SPACE",                 0x11},
	{"INQUIRY",               0x12},
	{"RECOVER_BUFFERED_DATA", 0x14},
	{"MODE_SELECT",           0x15},
	{"RESERVE",               0x16},
	{"RELEASE",               0x17},
	{"COPY",                  0x18},
	{"ERASE",                 0x19},
	{"MODE_SENSE",            0x1a},
	{"START_STOP",            0x1b},
	{"RECEIVE_DIAGNOSTIC",    0x1c},
	{"SEND_DIAGNOSTIC",       0x1d},
	{"ALLOW_MEDIUM_REMOVAL",  0x1e},
	{"SET_WINDOW",            0x24},
	{"READ_CAPACITY",         0x25},
	{"READ_10",               0x28},
	{"WRITE_10",              0x2a},
	{"SEEK_10",               0x2b},
	{"WRITE_VERIFY",          0x2e},
	{"VERIFY",                0x2f},
	{"SEARCH_HIGH",           0x30},
	{"SEARCH_EQUAL",          0x31},
	{"SEARCH_LOW",            0x32},
	{"SET_LIMITS",            0x33},
	{"PRE_FETCH",             0x34},
	{"READ_POSITION",         0x34},
	{"SYNCHRONIZE_CACHE",     0x35},
	{"LOCK_UNLOCK_CACHE",     0x36},
	{"READ_DEFECT_DATA",      0x37},
	{"MEDIUM_SCAN",           0x38},
	{"COMPARE",               0x39},
	{"COPY_VERIFY",           0x3a},
	{"WRITE_BUFFER",          0x3b},
	{"READ_BUFFER",           0x3c},
	{"UPDATE_BLOCK",          0x3d},
	{"READ_LONG",             0x3e},
	{"WRITE_LONG",            0x3f},
	{"CHANGE_DEFINITION",     0x40},
	{"WRITE_SAME",            0x41},
	{"READ_TOC",              0x43},
	{"LOG_SELECT",            0x4c},
	{"LOG_SENSE",             0x4d},
	{"MODE_SELECT_10",        0x55},
	{"RESERVE_10",            0x56},
	{"RELEASE_10",            0x57},
	{"MODE_SENSE_10",         0x5a},
	{"PERSISTENT_RESERVE_IN", 0x5e},
	{"PERSISTENT_RESERVE_OUT", 0x5f},
	{"MOVE_MEDIUM",           0xa5},
	{"READ_12",               0xa8},
	{"WRITE_12",              0xaa},
	{"WRITE_VERIFY_12",       0xae},
	{"SEARCH_HIGH_12",        0xb0},
	{"SEARCH_EQUAL_12",       0xb1},
	{"SEARCH_LOW_12",         0xb2},
	{"READ_ELEMENT_STATUS",   0xb8},
	{"SEND_VOLUME_TAG",       0xb6},
	{"WRITE_LONG_2",          0xea},
	{"(unknown)", -1},
};

/*
 * Inquiry
 */
static unsigned char inquiry_data[36] = {
	0x00 /* 0x20 for not-present */,
	0x80, 0x02, 0x02,
	0x1f
};

void sdm_fill_inquiry(Scsi_Cmnd *srb)
{
	char s[8+16+4 + 4];
	sprintf(s,"scsi-mem" /* 8b */
		"LM-DEMO (rubini)" /* 16b */
		"1.0 " /* 4b */);
	inquiry_data[0] = 0x00 /* present ? 0x00 : 0x20 */;
	memcpy(inquiry_data+8, s, 8+16+4);
	memset(srb->request_buffer, 0, srb->request_bufflen);
	memcpy(srb->request_buffer, inquiry_data, 36);
}

/*
 * Command dispatcher, high level
 */
void sdm_scsi_command(struct sdm_dev *dev, Scsi_Cmnd *srb)
{
	struct sdm_scsi_cmdname *p;
	
	for (p = sdm_scsi_cmdtable; p->value >= 0; p++)
		if (srb->cmnd[0] == p->value)
			break;
	if (sdm_verbose)
	    printk("sdm: \"%s\" len %i dir %i (id %i, lun %i)\n",p->name,
		   srb->cmd_len, srb->sc_data_direction,
		   srb->device->id, srb->device->lun);

	/* we are not multi-target */
	if (srb->device->id) {
	    srb->result = DID_BAD_TARGET << 16;
	    return;
	}

	if (srb->sc_data_direction == SCSI_DATA_UNKNOWN) {
		printk("sdm: unknown data direction\n");
		srb->result = DID_ERROR << 16;
		return;
	}
	
	/* we can't do a START_STOP */
	if (srb->cmnd[0] == START_STOP) {
		printk("sdm: ignoring start_stop\n");
		srb->result = SAM_STAT_GOOD;
		return;
	}

	/* inquiry had the same code for present and absent devices */
	if(srb->cmnd[0] == INQUIRY) {
		sdm_fill_inquiry(srb);
		srb->result = SAM_STAT_GOOD;
		return;
	}

	sdm_scsi_lowlevel_command(srb);
	return;
}


/* We need a thread to run SCSI commands */
static int sdm_thread(void *arg)
{
	struct sdm_dev *dev = (struct sdm_dev *)arg;

	daemonize("sdm-thread");
	current->flags |= PF_NOFREEZE;

	complete(&dev->notify); /* we are running */
	
	while (1) {
		/* sleep */
                if(down_interruptible(&dev->sema))
                        break;

		/* check if we should exit */
		if (dev->doexit) break;

		/* run command (possibly printk it too) */
		sdm_scsi_command(dev, dev->srb);

		/* tell we are done */
		dev->srb->scsi_done(dev->srb);
	}
	complete_and_exit(&dev->notify, 0);
}

/* This is called when a new request comes in */
static int sdm_queuecommand(Scsi_Cmnd *srb, void (*done)(Scsi_Cmnd *))
{
	srb->scsi_done = done;
	sdm_dev.srb = srb;
	up(&sdm_dev.sema);
	return 0;
}

static const char* sdm_host_info(struct Scsi_Host *host)
{
        return "SCSI emulation for RAM disk";
}

/* The template for our host */
struct scsi_host_template sdm_host_template = {
        /* basic userland interface stuff */
        .name =                         "scsi-disk-memory",
        .proc_name =                    "sdm",
        .info =                         sdm_host_info,

        /* command interface -- queued only */
        .queuecommand =                 sdm_queuecommand,

        /* queue commands only, only one command per LUN */
        .can_queue =                    1,
        .cmd_per_lun =                  1,

        /* unknown initiator id */
        .this_id =                      -1,

        /* lots of sg segments can be handled */
        .sg_tablesize =                 SG_ALL,

        /* easy: max one page */
        .max_sectors =                  PAGE_SIZE/512,

        .use_clustering =               FALSE, /* safer */

        /* emulated HBA */
        .emulated =                     TRUE,

        /* we do our own delay after a device or bus reset */
        .skip_settle_delay =            1,

        /* sysfs device attributes */
        .sdev_attrs =                   NULL /* sysfs_device_attr_list */,

        /* module management */
        .module =                       THIS_MODULE
};


static int sdm_init(void)
{
        init_MUTEX_LOCKED(&sdm_dev.sema);
        init_completion(&sdm_dev.notify);
	sdm_dev.host = scsi_host_alloc(&sdm_host_template, 0);
	if (!sdm_dev.host) return -EBUSY;
	kernel_thread(sdm_thread, &sdm_dev, CLONE_VM);
        /* Wait for the thread to start */
        wait_for_completion(&sdm_dev.notify);
	scsi_add_host(sdm_dev.host, NULL);
	scsi_scan_host(sdm_dev.host);
	return 0;
}

static void sdm_exit(void)
{
	scsi_remove_host(sdm_dev.host);
	sdm_dev.doexit = 1;
	up(&sdm_dev.sema);
	wait_for_completion(&sdm_dev.notify);
	scsi_host_put(sdm_dev.host);
}
module_init(sdm_init);
module_exit(sdm_exit);
